"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sift_1 = require("sift");
const log = require("./log");
const db = require("./db");
const _1 = require("./_");
const Event_1 = require("./Event");
const queue_1 = require("./queue");
const ProgramItem_1 = require("./ProgramItem");
class Program {
    constructor() {
        this._itemMap = new Map();
        this._programGCInterval = _1.default.config.server.programGCInterval || 1000 * 60 * 15;
        this._load();
        setTimeout(this._gc.bind(this), this._programGCInterval);
    }
    static get(id) {
        return _1.default.program.get(id);
    }
    static exists(id) {
        return _1.default.program.exists(id);
    }
    static findByQuery(query) {
        return _1.default.program.findByQuery(query);
    }
    static findByServiceId(serviceId) {
        return _1.default.program.findByServiceId(serviceId);
    }
    static all() {
        return _1.default.program.items;
    }
    /** CAUTION: This getter method creates a new Array object every time. */
    get items() {
        return Array.from(this._itemIterator);
    }
    get _itemIterator() {
        return this._itemMap.values();
    }
    add(item, firstAdd = false) {
        if (this.exists(item.id)) {
            return;
        }
        const removedIds = [];
        if (firstAdd === false) {
            _1.default.program.findConflicts(item.data.networkId, item.data.serviceId, item.data.startAt, item.data.startAt + item.data.duration).forEach(conflictedItem => {
                this.remove(conflictedItem);
                log.debug("ProgramItem#%d (networkId=%d, eventId=%d) has removed for redefine to ProgramItem#%d (eventId=%d)", conflictedItem.data.id, conflictedItem.data.networkId, conflictedItem.data.eventId, item.data.id, item.data.eventId);
                removedIds.push(conflictedItem.data.id);
            });
        }
        this._itemMap.set(item.id, item);
        if (firstAdd === false) {
            Event_1.default.emit("program", "create", item.data);
            removedIds.forEach(id => Event_1.default.emit("program", "redefine", { from: id, to: item.data.id }));
        }
        this.save();
    }
    get(id) {
        return this._itemMap.get(id) || null;
    }
    remove(item) {
        if (this._itemMap.delete(item.id)) {
            this.save();
        }
    }
    exists(id) {
        return this._itemMap.has(id);
    }
    findByQuery(query) {
        // Pass `this.items` instead of `this._itemIterator`.
        // Because IterableIterator<T> doesn't have the `filter()` method.
        return sift_1.default(query, this.items);
    }
    findByServiceId(serviceId) {
        const items = [];
        for (const item of this._itemIterator) {
            if (item.data.serviceId === serviceId) {
                items.push(item);
            }
        }
        return items;
    }
    findConflicts(networkId, serviceId, start, end) {
        const items = [];
        for (const item of this._itemIterator) {
            if (item.data.networkId === networkId &&
                item.data.serviceId === serviceId &&
                item.data.startAt >= start &&
                item.data.startAt < end) {
                items.push(item);
            }
        }
        return items;
    }
    findByNetworkIdAndReplace(networkId, programs) {
        let count = 0;
        // The `reverse()` method never changes the original data.  Because
        // `this.items` returns a new Array object created from the original
        // data.
        for (const item of this.items.reverse()) {
            if (item.data.networkId === networkId) {
                // Calling `this.remove(item)` here is safe.  Because that never
                // changes the Array object we're iterating here.
                this.remove(item);
                --count;
            }
        }
        for (const program of programs) {
            this.add(new ProgramItem_1.default(program), true);
            ++count;
        }
        log.debug("programs replaced (networkId=%d, count=%d)", networkId, count);
        this.save();
    }
    save() {
        clearTimeout(this._saveTimerId);
        this._saveTimerId = setTimeout(() => this._save(), 3000);
    }
    _load() {
        log.debug("loading programs...");
        const now = Date.now();
        let dropped = false;
        db.loadPrograms(_1.default.configIntegrity.channels).forEach(program => {
            if (program.networkId === undefined) {
                dropped = true;
                return;
            }
            if (now > (program.startAt + program.duration)) {
                dropped = true;
                return;
            }
            this.add(new ProgramItem_1.default(program), true);
        });
        if (dropped) {
            this.save();
        }
    }
    _save() {
        log.debug("saving programs...");
        db.savePrograms(this._collectDBData(), _1.default.configIntegrity.channels);
    }
    _collectDBData() {
        const results = [];
        for (const item of this._itemIterator) {
            results.push(item.data);
        }
        return results;
    }
    _gc() {
        log.debug("Program GC has queued");
        queue_1.default.add(async () => {
            const now = Date.now();
            let count = 0;
            for (const program of this._itemIterator) {
                if (now > (program.data.startAt + program.data.duration)) {
                    ++count;
                    this.remove(program);
                }
            }
            setTimeout(this._gc.bind(this), this._programGCInterval);
            log.info("Program GC has finished and removed %d programs", count);
        });
    }
}
exports.default = Program;
//# sourceMappingURL=Program.js.map