"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
   Copyright 2016 kanreisa

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
const child_process = require("child_process");
const common = require("./common");
const log = require("./log");
const _1 = require("./_");
const status_1 = require("./status");
const TunerDevice_1 = require("./TunerDevice");
const TSFilter_1 = require("./TSFilter");
class Tuner {
    constructor() {
        this._devices = [];
        this._load();
    }
    static all() {
        return _1.default.tuner.devices;
    }
    static get(index) {
        return _1.default.tuner.get(index);
    }
    static typeExists(type) {
        return _1.default.tuner.typeExists(type);
    }
    static getChannelStream(channel, user) {
        return _1.default.tuner.getChannelStream(channel, user);
    }
    static getServiceStream(service, user) {
        return _1.default.tuner.getServiceStream(service, user);
    }
    static getProgramStream(program, user) {
        return _1.default.tuner.getProgramStream(program, user);
    }
    static getEPG(channel, time) {
        return _1.default.tuner.getEPG(channel, time);
    }
    static getServices(channel) {
        return _1.default.tuner.getServices(channel);
    }
    get devices() {
        return this._devices;
    }
    get(index) {
        const l = this._devices.length;
        for (let i = 0; i < l; i++) {
            if (this._devices[i].index === index) {
                return this._devices[i];
            }
        }
        return null;
    }
    typeExists(type) {
        const l = this._devices.length;
        for (let i = 0; i < l; i++) {
            if (this._devices[i].config.types.indexOf(type) !== -1) {
                return true;
            }
        }
        return false;
    }
    getChannelStream(channel, user) {
        let networkId;
        const services = channel.getServices();
        if (services.length !== 0) {
            networkId = services[0].networkId;
        }
        const setting = {
            channel: channel,
            networkId: networkId,
            parseEIT: true
        };
        return this._getStream(setting, user);
    }
    getServiceStream(service, user) {
        const setting = {
            channel: service.channel,
            serviceId: service.serviceId,
            networkId: service.networkId,
            parseEIT: true
        };
        return this._getStream(setting, user);
    }
    getProgramStream(program, user) {
        const setting = {
            channel: program.service.channel,
            serviceId: program.data.serviceId,
            eventId: program.data.eventId,
            networkId: program.data.networkId,
            parseEIT: true
        };
        return this._getStream(setting, user);
    }
    async getEPG(channel, time) {
        if (!time) {
            time = _1.default.config.server.epgRetrievalTime || 1000 * 60 * 10;
        }
        let networkId;
        const services = channel.getServices();
        if (services.length !== 0) {
            networkId = services[0].networkId;
        }
        const setting = {
            channel: channel,
            networkId: networkId,
            noProvide: true,
            parseEIT: true
        };
        const user = {
            id: "Mirakurun:getEPG()",
            priority: -1,
            disableDecoder: true
        };
        const stream = await this._getStream(setting, user);
        if (stream === null) {
            return Promise.resolve();
        }
        else {
            return new Promise((resolve) => {
                setTimeout(() => stream.emit("close"), time);
                stream.once("epgReady", () => stream.emit("close"));
                stream.once("close", resolve);
            });
        }
    }
    async getServices(channel) {
        const setting = {
            channel: channel,
            noProvide: true,
            parseNIT: true,
            parseSDT: true
        };
        const user = {
            id: "Mirakurun:getServices()",
            priority: -1,
            disableDecoder: true
        };
        const stream = await this._getStream(setting, user);
        return new Promise((resolve, reject) => {
            let network = {
                networkId: -1,
                areaCode: -1,
                remoteControlKeyId: -1
            };
            let services = null;
            setTimeout(() => stream.emit("close"), 20000);
            Promise.all([
                new Promise((resolve, reject) => {
                    stream.once("network", _network => {
                        network = _network;
                        resolve();
                    });
                }),
                new Promise((resolve, reject) => {
                    stream.once("services", _services => {
                        services = _services;
                        resolve();
                    });
                })
            ]).then(() => stream.emit("close"));
            stream.once("close", () => {
                stream.removeAllListeners("network");
                stream.removeAllListeners("services");
                if (network.networkId === -1) {
                    reject(new Error("stream has closed before get network"));
                }
                else if (services === null) {
                    reject(new Error("stream has closed before get services"));
                }
                else {
                    if (network.remoteControlKeyId !== -1) {
                        services.forEach(service => {
                            service.remoteControlKeyId = network.remoteControlKeyId;
                        });
                    }
                    resolve(services);
                }
            });
        });
    }
    _load() {
        log.debug("loading tuners...");
        const tuners = _1.default.config.tuners;
        tuners.forEach((tuner, i) => {
            if (!tuner.name || !tuner.types || (!tuner.remoteMirakurunHost && !tuner.command)) {
                log.error("missing required property in tuner#%s configuration", i);
                return;
            }
            if (typeof tuner.name !== "string") {
                log.error("invalid type of property `name` in tuner#%s configuration", i);
                return;
            }
            if (Array.isArray(tuner.types) === false) {
                console.log(tuner);
                log.error("invalid type of property `types` in tuner#%s configuration", i);
                return;
            }
            if (!tuner.remoteMirakurunHost && typeof tuner.command !== "string") {
                log.error("invalid type of property `command` in tuner#%s configuration", i);
                return;
            }
            if (tuner.dvbDevicePath && typeof tuner.dvbDevicePath !== "string") {
                log.error("invalid type of property `dvbDevicePath` in tuner#%s configuration", i);
                return;
            }
            if (tuner.remoteMirakurunHost && typeof tuner.remoteMirakurunHost !== "string") {
                log.error("invalid type of property `remoteMirakurunHost` in tuner#%s configuration", i);
                return;
            }
            if (tuner.remoteMirakurunPort && Number.isInteger(tuner.remoteMirakurunPort) === false) {
                log.error("invalid type of property `remoteMirakurunPort` in tuner#%s configuration", i);
                return;
            }
            if (tuner.remoteMirakurunDecoder !== undefined && typeof tuner.remoteMirakurunDecoder !== "boolean") {
                log.error("invalid type of property `remoteMirakurunDecoder` in tuner#%s configuration", i);
                return;
            }
            if (tuner.isDisabled) {
                return;
            }
            this._devices.push(new TunerDevice_1.default(i, tuner));
        });
        log.info("%s of %s tuners loaded", this._devices.length, tuners.length);
        return this;
    }
    _getStream(setting, user) {
        return new Promise((resolve, reject) => {
            const devices = this._getDevicesByType(setting.channel.type);
            let tryCount = 50;
            const length = devices.length;
            function find() {
                let device = null;
                // 1. join to existing
                for (let i = 0; i < length; i++) {
                    if (devices[i].isAvailable === true && devices[i].channel === setting.channel) {
                        device = devices[i];
                        break;
                    }
                }
                // x. use remote data
                if (device === null && setting.noProvide === true) {
                    const remoteDevice = devices.find(device => device.isRemote);
                    if (remoteDevice) {
                        if (setting.networkId !== undefined && setting.parseEIT === true) {
                            remoteDevice.getRemotePrograms({ networkId: setting.networkId })
                                .then(async (programs) => {
                                await common.sleep(1000);
                                _1.default.program.findByNetworkIdAndReplace(setting.networkId, programs);
                                await common.sleep(1000);
                            })
                                .then(() => resolve(null))
                                .catch(err => reject(err));
                            return;
                        }
                    }
                }
                // 2. start as new
                if (device === null) {
                    for (let i = 0; i < length; i++) {
                        if (devices[i].isFree === true) {
                            device = devices[i];
                            break;
                        }
                    }
                }
                // 3. replace existing
                if (device === null) {
                    for (let i = 0; i < length; i++) {
                        if (devices[i].isAvailable === true && devices[i].users.length === 0) {
                            device = devices[i];
                            break;
                        }
                    }
                }
                // 4. takeover existing
                if (device === null) {
                    devices.sort((t1, t2) => {
                        return t1.getPriority() - t2.getPriority();
                    });
                    for (let i = 0; i < length; i++) {
                        if (devices[i].isUsing === true && devices[i].getPriority() < user.priority) {
                            device = devices[i];
                            break;
                        }
                    }
                }
                if (device === null) {
                    --tryCount;
                    if (tryCount > 0) {
                        setTimeout(find, 250);
                    }
                    else {
                        reject(new Error("no available tuners"));
                    }
                }
                else {
                    const tsFilter = new TSFilter_1.default({
                        networkId: setting.networkId,
                        serviceId: setting.serviceId,
                        eventId: setting.eventId,
                        noProvide: setting.noProvide,
                        parseNIT: setting.parseNIT,
                        parseSDT: setting.parseSDT,
                        parseEIT: setting.parseEIT
                    });
                    device.startStream(user, tsFilter, setting.channel)
                        .then(() => {
                        if (user.disableDecoder === true || device.decoder === null) {
                            resolve(tsFilter);
                        }
                        else {
                            const decoder = child_process.spawn(device.decoder);
                            ++status_1.default.streamCount.decoder;
                            decoder.stderr.pipe(process.stderr);
                            decoder.stdout.once("close", () => {
                                tsFilter.emit("close");
                                --status_1.default.streamCount.decoder;
                            });
                            decoder.stdin.once("finish", () => decoder.kill("SIGKILL"));
                            tsFilter.once("close", () => decoder.stdin.end());
                            tsFilter.pipe(decoder.stdin);
                            resolve(decoder.stdout);
                        }
                    })
                        .catch((err) => {
                        tsFilter.emit("close");
                        reject(err);
                    });
                }
            }
            find();
        });
    }
    _getDevicesByType(type) {
        const devices = [];
        const l = this._devices.length;
        for (let i = 0; i < l; i++) {
            if (this._devices[i].config.types.indexOf(type) !== -1) {
                devices.push(this._devices[i]);
            }
        }
        return devices;
    }
}
exports.default = Tuner;
//# sourceMappingURL=Tuner.js.map